/*
 * PaginatedGUI.java
 * Copyright (c) 2017 Sam Jakob Harker
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

package str.playerwarps.gui.types;

import net.md_5.bungee.api.ChatColor;
import str.playerwarps.PlayerWarps;
import str.playerwarps.commands.gui.GUI;
import str.playerwarps.gui.ItemBuilder;
import str.playerwarps.gui.buttons.GUIButton;
import str.playerwarps.gui.buttons.InventoryListenerGUI;
import str.playerwarps.util.TitleUpdater_1_8_R3;
import str.playerwarps.util.Utils;

import org.bukkit.Bukkit;
import org.bukkit.Material;
import org.bukkit.OfflinePlayer;
import org.bukkit.entity.HumanEntity;
import org.bukkit.entity.Player;
import org.bukkit.inventory.Inventory;
import org.bukkit.inventory.InventoryHolder;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.ItemMeta;
import org.bukkit.plugin.java.JavaPlugin;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.regex.Pattern;

public class PaginatedGUI implements InventoryHolder {

    /* BEGIN: CONFIGURATION */
    private static final String CHAT_PREFIX = "&6";
    private static final String NO_PREVIOUS_PAGES = "&cThere are no previous pages.";
    private static final String NO_ADDITIONAL_PAGES = "&cThere are no additional pages.";

    private static final String PREVIOUS_PAGE = "&cPrevious Page &7(&6{previousPage}&7/&6{maxPages}&7)";
    private static final String CURRENT_PAGE = "&aPage &7(&6{currentPage}&7/&6{maxPages}&7)";
    private static final String NEXT_PAGE = "&aNext Page &7(&6{nextPage}&7/&6{maxPages}&7)";
    public static final String OWN_WARP = "&a&lYour Player Warp";
    /* END: CONFIGURATION */

    private static InventoryListenerGUI inventoryListenerGUI;
    public static HashMap<Player, Boolean> inSubMenu = new HashMap<Player, Boolean>();
    private Map<Integer, GUIButton> items;
    private Map<Integer, GUIButton> toolbarItems;
    private Map<Integer, GUIButton> glassBarItems;
    private int currentPage;
    private String name;

    /**
     * Creates a PaginatedGUI. This is a Spigot 'Inventory Menu' that
     * will automatically add pages to accommodate additional items.
     *
     * <br>
     *
     * Color Codes are supported (and should be prepended with an
     * ampersand [&amp;]; e.g. &amp;c for red.)
     *
     * @param name The desired name of the PaginatedGUI.
     */
    public PaginatedGUI(String name){
        items = new HashMap<>();
        toolbarItems = new HashMap<>();
        glassBarItems = new HashMap<>();
        currentPage = 0;
        this.name = ChatColor.translateAlternateColorCodes('&', name);
    }

    /**
     * Sets the display name of the PaginatedGUI.
     *
     * <br>
     *
     * Color Codes are supported (and should be prepended with an
     * ampersand [&amp;]; e.g. &amp;c for red.)
     *
     * @param name The desired name of the PaginatedGUI.
     */
    public void setDisplayName(String name){
        this.name = ChatColor.translateAlternateColorCodes('&', name);
    }

    /**
     * Gets the display name of the PaginatedGUI.
     *
     * <br>
     *
     * <b>Note:</b> If your inventory's display name contains
     * color codes, this will have substituted the
     * ampersands (&amp;)s with the rendering engine's
     * symbol (&sect;).
     *
     * @return The inventory's display name.
     */
    public String getDisplayName(){
        return name;
    }

    /**
     * Adds the provided {@link GUIButton} to the PaginatedGUI.
     *
     * <br>
     *
     * <b>Note:</b> This will place the button after the highest slot.
     * So if you have buttons in slot 0, 1 and 5, this will place the
     * added button in slot 6.
     *
     * @param button The button you wish to add.
     */
    public void addButton(GUIButton button) {
        // Get the current maximum slot in the 'items' list.
        int slot = 0;
        for (int nextSlot : items.keySet()) {
            if (nextSlot > slot) {
                slot = nextSlot;
            }
        }
        if (items.keySet().size() == 0) {
            // Put the button in that slot.
            items.put(slot, button);
        }
        else {
	        // Add one to get the next maximum slot.
	        slot++;
	        
	        // Put the button in that slot.
	        items.put(slot, button);
        }
    }
    
    public void addBarButton(GUIButton button) {
        // Get the current maximum slot in the 'items' list.
        int slot = 0;
        for (int nextSlot : glassBarItems.keySet()) {
            if (nextSlot > slot) {
                slot = nextSlot;
            }
        }

        // Add one to get the next maximum slot.
        slot++;

        // Put the button in that slot.
        glassBarItems.put(slot, button);
    }

    /**
     * Adds the provided {@link GUIButton} but places it in the desired slot in the PaginatedGUI.
     *
     * @param slot The desired slot for the button.
     * @param button The button you wish to add.
     */
    public void setButton(int slot, GUIButton button){
        items.put(slot, button);
    }

    /**
     * Removes the {@link GUIButton} from the provided slot.
     *
     * @param slot The slot containing the button you wish to remove.
     */
    public void removeButton(int slot){
        items.remove(slot);
    }

    /**
     * Gets the {@link GUIButton} in the provided slot.
     *
     * @param slot The slot containing the GUIButton that you wish to get.
     * @return The GUIButton in the provided slot.
     */
    public GUIButton getButton(int slot) {
        if (slot < 36) {
            if (currentPage > 0) {
                return items.get(slot + (36 * currentPage));
            }
            return items.get(slot);
        }
        else {
            return toolbarItems.get(slot - 36);
        }
    }

    /**
     * Adds the provided {@link GUIButton} but places it in the desired slot in the PaginatedGUI's toolbar.
     *
     * @param slot The desired slot for the button.
     * @param button The button you wish to add.
     * @throws IllegalArgumentException This will occur if the slot is less than 0 or higher than 8 (as this is outside the toolbar slot range.)
     */
    public void setToolbarItem(int slot, GUIButton button) {
        if(slot < 0 || slot > 8){
            throw new IllegalArgumentException("The desired slot is outside the bounds of the toolbar slot range. [0-8]");
        }

        toolbarItems.put(slot, button);
    }

    /**
     * Removes the {@link GUIButton} from the provided slot.
     *
     * @param slot The slot containing the button you wish to remove.
     * @throws IllegalArgumentException This will occur if the slot is less than 0 or higher than 8 (as this is outside the toolbar slot range.)
     */
    public void removeToolbarItem(int slot) {
        if(slot < 0 || slot > 8){
            throw new IllegalArgumentException("The desired slot is outside the bounds of the toolbar slot range. [0-8]");
        }

        toolbarItems.remove(slot);
    }

    /**
     * Increments the current page.
     * You will need to refresh the inventory for those who have it open with {@link #refreshInventory(HumanEntity)}
     *
     * @return Whether or not the page could be changed (false when the max page is currently open as it cannot go further.)
     */
    public boolean nextPage(){
        if(currentPage < getFinalPage()){
            currentPage++;
            return true;
        }else{
            return false;
        }
    }
    /**
     * Decrements the current page.
     * You will need to refresh the inventory for those who have it open with {@link #refreshInventory(HumanEntity)}
     *
     * @return Whether or not the page could be changed (false when the first page is currently active as it cannot go further.)
     */

    public boolean previousPage(){
        if(currentPage > 0) {
            currentPage--;
            return true;
        }else{
            return false;
        }
    }

    /**
     * An alias for {@link #getFinalPage()}.
     *
     * @deprecated Use {@link #getFinalPage()} instead.
     * @return The highest page number that can be viewed.
     */
    public int getMaxPage(){
        return getFinalPage();
    }

    /**
     * Gets the number of the final page of the PaginatedGUI.
     *
     * @return The highest page number that can be viewed.
     */
    public int getFinalPage(){
        // Get the highest slot number.
        int slot = 0;
        for(int nextSlot : items.keySet()){
            if(nextSlot > slot){
                slot = nextSlot;
            }
        }

        // Add one to make the math easier.
        double highestSlot = slot + 1;

        // Divide by 45 and round up to get the page number.
        // Then subtract one to convert it to an index.
        return (int) Math.ceil(highestSlot / (double) 36) - 1;
    }
    
    /**
     * Gets the number of the current page of the PaginatedGUI.
     *
     * @return The current page number.
     */
    public int getCurrentPage(){
        return currentPage;
    }
    
    /**
     * Sets the number of the current page of the PaginatedGUI.
     *
     * @return Sets the current page number to what you wish.
     */
    public void setCurrentPage(int page){
    	currentPage = page;
    }

    /**
     * Simply an alias that executes {@link HumanEntity#closeInventory()} and then
     * {@link HumanEntity#openInventory(Inventory)}.
     *
     * @param holder The HumanEntity that you wish to refresh the inventory for.
     */
    public void refreshInventory(HumanEntity holder) {
        Inventory items = getInventory();
        holder.getOpenInventory().getTopInventory().setContents(items.getContents());
    }

    /**
     * Returns the Spigot {@link Inventory} that represents the PaginatedGUI.
     * This can then by shown to a player using {@link HumanEntity#openInventory(Inventory)}.
     *
     * <br>
     *
     * This also allows getting the PaginatedGUI instance with {@link InventoryHolder#getInventory()}.
     * Used internally ({@link InventoryListenerGUI}) to get the GUIButton and therefore listener from the raw slot.
     *
     * @return The Spigot Inventory that represents the PaginatedGUI.
     */
    @Override
    public Inventory getInventory() {
        // Create an inventory (and set an appropriate size.)
        // TODO: Allow customisation of inventory size. Maybe at first, only if the inventory is not paginated.
        Inventory inventory = Bukkit.createInventory(this, (getFinalPage() > 0) ? 45 : 36, name);

        /* BEGIN PAGINATION */
        GUIButton backButton = new GUIButton(ItemBuilder.start(Material.MELON).name(PREVIOUS_PAGE
        		.replaceAll(Pattern.quote("{previousPage}"), String.valueOf((currentPage + 1) - 1))
        		.replaceAll(Pattern.quote("{maxPages}"), String.valueOf(getFinalPage() + 1))
        		)
        		.build());
        GUIButton nextButton = new GUIButton(ItemBuilder.start(Material.SPECKLED_MELON).name(NEXT_PAGE
        		.replaceAll(Pattern.quote("{nextPage}"), String.valueOf((currentPage + 1) + 1))
        		.replaceAll(Pattern.quote("{maxPages}"), String.valueOf(getFinalPage() + 1))
        		)
        		.build());
        GUIButton glassPane = new GUIButton(ItemBuilder.start(Material.STAINED_GLASS_PANE).data((short)7).name(" ").build());
        glassPane.setListener(event -> {
			event.setCancelled(true);
		});
        if (getFinalPage() > 0) {
        	List<String> lore = new ArrayList<String>();
        	lore.add(ChatColor.translateAlternateColorCodes('&', "&eRight click to view all pages"));
        	
            ItemMeta stackMeta1 = backButton.getItem().getItemMeta();
            stackMeta1.setLore(lore);
            backButton.getItem().setItemMeta(stackMeta1);
            
            ItemMeta stackMeta2 = nextButton.getItem().getItemMeta();
            stackMeta2.setLore(lore);
            nextButton.getItem().setItemMeta(stackMeta2);
        }
        backButton.setListener(event -> {
            event.setCancelled(true);
            PaginatedGUI menu = (PaginatedGUI) event.getClickedInventory().getHolder();
            if (event.isLeftClick()) {
            	if(!menu.previousPage()){
            		event.getWhoClicked().sendMessage(ChatColor.translateAlternateColorCodes('&',
            				CHAT_PREFIX + NO_PREVIOUS_PAGES));
            		return;
            	}
            	
            	refreshInventory(event.getWhoClicked());
            	Player player = (Player) event.getWhoClicked();
            	TitleUpdater_1_8_R3.update(player, ChatColor.translateAlternateColorCodes('&', "Player Warps &7(&6" + (menu.getCurrentPage() + 1) + "/" + (menu.getFinalPage() + 1) + "&7)"));
            }
            else if (event.isRightClick()) {
            	PaginatedGUIAllPages menu2 = new PaginatedGUIAllPages("Open a page...");
            	menu2.setCloseListener(event2 -> {
    				Bukkit.getScheduler().runTaskLater(PlayerWarps.get(), new Runnable() {
    					public void run() {
    						Player player = (Player) event2.getPlayer();
    						if (inSubMenu.get(player) == true) {
    							inSubMenu.remove(player);
    							Utils.performCommand(player, "playerwarps page " + (currentPage + 1));
    						}
    					}
    				}, 1L);
    			});
            	for (int i = 0; i < (getFinalPage() + 1); i++) {
            		if (currentPage == i) {
                    	List<String> lore = new ArrayList<String>();
                    	lore.add(ChatColor.translateAlternateColorCodes('&', "&7Click to open"));
                		lore.add("");
                		lore.add(ChatColor.translateAlternateColorCodes('&', "&aCurrent page"));
                		GUIButton book = new GUIButton(ItemBuilder.start(Material.ENCHANTED_BOOK).amount(i + 1).lore(lore).name("&ePage " + (i + 1)).build());
                		final int o = i;
                		book.setListener(event1 -> {
                			event1.setCancelled(true);
                			Player player = (Player) event1.getWhoClicked();
                			if (player.getOpenInventory() != null) player.closeInventory();
                			if (inSubMenu.containsKey(player)) {
                				inSubMenu.put(player, false);
                			}
                			Utils.performCommand(player, "playerwarp page " + (o + 1));
                		});
                		menu2.addButton(book);
                	}
                	else {
                    	List<String> lore = new ArrayList<String>();
                    	lore.add(ChatColor.translateAlternateColorCodes('&', "&7Click to open"));
                		GUIButton book = new GUIButton(ItemBuilder.start(Material.BOOK).amount(i + 1).lore(lore).name("&ePage " + (i + 1)).build());
                		final int o = i;
                		book.setListener(event1 -> {
                			event1.setCancelled(true);
                			Player player = (Player) event1.getWhoClicked();
                			if (player.getOpenInventory() != null) player.closeInventory();
                			if (inSubMenu.containsKey(player)) {
                				inSubMenu.put(player, false);
                			}
                			Utils.performCommand(player, "playerwarp page " + (o + 1));
                		});
                		menu2.addButton(book);
                	}
            	}
            	Player player = (Player) event.getWhoClicked();
            	if (event.getWhoClicked().getOpenInventory() != null) event.getWhoClicked().closeInventory();
            	if (inSubMenu.containsKey(player)) {
            		inSubMenu.remove(player);
            		inSubMenu.put(player, true);
            	}
            	else {
            		inSubMenu.put(player, true);
            	}
            	event.getWhoClicked().openInventory(menu2.getInventory());
            }
        });

        nextButton.setListener(event -> {
            event.setCancelled(true);
            PaginatedGUI menu = (PaginatedGUI) event.getClickedInventory().getHolder();
            if (event.isLeftClick()) {
            	if(!menu.nextPage()){
            		event.getWhoClicked().sendMessage(ChatColor.translateAlternateColorCodes('&',
            				CHAT_PREFIX + NO_ADDITIONAL_PAGES));
            		return;
            	}
            	
            	refreshInventory(event.getWhoClicked());
            	Player player = (Player) event.getWhoClicked();
            	TitleUpdater_1_8_R3.update(player, ChatColor.translateAlternateColorCodes('&', "Player Warps &7(&6" + (menu.getCurrentPage() + 1) + "/" + (menu.getFinalPage() + 1) + "&7)"));
            }
            if (event.isRightClick()) {
            	PaginatedGUIAllPages menu2 = new PaginatedGUIAllPages("Open a page...");
            	menu2.setCloseListener(event2 -> {
    				Bukkit.getScheduler().runTaskLater(PlayerWarps.get(), new Runnable() {
    					public void run() {
    						Player player = (Player) event2.getPlayer();
    						if (inSubMenu.get(player) == true) {
    							inSubMenu.remove(player);
    							Utils.performCommand(player, "playerwarps page " + (currentPage + 1));
    						}
    					}
    				}, 1L);
    			});
            	for (int i = 0; i < (getFinalPage() + 1); i++) {
                	if (currentPage == i) {
                    	List<String> lore = new ArrayList<String>();
                    	lore.add(ChatColor.translateAlternateColorCodes('&', "&7Click to open"));
                		lore.add("");
                		lore.add(ChatColor.translateAlternateColorCodes('&', "&aCurrent page"));
                		GUIButton book = new GUIButton(ItemBuilder.start(Material.ENCHANTED_BOOK).amount(i + 1).lore(lore).name("&ePage " + (i + 1)).build());
                		final int o = i;
                		book.setListener(event1 -> {
                			event1.setCancelled(true);
                			Player player = (Player) event1.getWhoClicked();
                			if (player.getOpenInventory() != null) player.closeInventory();
                			if (inSubMenu.containsKey(player)) {
                				inSubMenu.put(player, false);
                			}
                			Utils.performCommand(player, "playerwarp page " + (o + 1));
                		});
                		menu2.addButton(book);
                	}
                	else {
                    	List<String> lore = new ArrayList<String>();
                    	lore.add(ChatColor.translateAlternateColorCodes('&', "&7Click to open"));
                		GUIButton book = new GUIButton(ItemBuilder.start(Material.BOOK).amount(i + 1).lore(lore).name("&ePage " + (i + 1)).build());
                		final int o = i;
                		book.setListener(event1 -> {
                			event1.setCancelled(true);
                			Player player = (Player) event1.getWhoClicked();
                			if (player.getOpenInventory() != null) player.closeInventory();
                			if (inSubMenu.containsKey(player)) {
                				inSubMenu.put(player, false);
                			}
                			Utils.performCommand(player, "playerwarp page " + (o + 1));
                		});
                		menu2.addButton(book);
                	}
            	}
            	if (event.getWhoClicked().getOpenInventory() != null) event.getWhoClicked().closeInventory();
            	Player player = (Player) event.getWhoClicked();
            	if (inSubMenu.containsKey(player)) {
            		inSubMenu.remove(player);
            		inSubMenu.put(player, true);
            	}
            	else {
            		inSubMenu.put(player, true);
            	}
            	event.getWhoClicked().openInventory(menu2.getInventory());
            }
        });
        ItemStack glass = new ItemStack(Material.STAINED_GLASS_PANE);
        ItemMeta itemMeta = glass.getItemMeta();
        itemMeta.setDisplayName(" ");
        glass.setDurability((short)7);
        glass.setItemMeta(itemMeta);
        /* END PAGINATION */

        // Where appropriate, include pagination.
        if(currentPage > 0) {
            toolbarItems.put(0, backButton);
        }
        if (currentPage == 0) {
        	if (toolbarItems.get(0) != null) {
        		toolbarItems.remove(0);
        	}
        }
        if(currentPage < getFinalPage())
            toolbarItems.put(8, nextButton);
        if (currentPage == getFinalPage()) {
        	if (toolbarItems.get(8) != null) {
        		toolbarItems.remove(8);
        	}
        }

        // Add the main inventory items
        int counter = 0;
        for(int key = (currentPage * 36); key <= Collections.max(items.keySet()); key++){
            if(counter >= 36)
                break;

            if(items.containsKey(key)) {
                inventory.setItem(counter, items.get(key).getItem());
            }

            counter++;
        }
        
        if (currentPage < getFinalPage()) {
	        if (currentPage == 0 && currentPage == getFinalPage()) {
	        	for (int i = 0; i < 9; i++) {
	        		toolbarItems.put(i, glassPane);
	        	}
	        }
	        else if (currentPage > 0 && currentPage != getFinalPage()) {
	        	for (int i = 1; i < 8; i++) {
	        		toolbarItems.put(i, glassPane);
	        	}
	        }
	        else if (currentPage == 0) {
	        	for (int i = 0; i < 8; i++) {
	        		toolbarItems.put(i, glassPane);
	        	}
	        }
	        else if (currentPage == getFinalPage()) {
	        	for (int i = 1; i < 9; i++) {
	        		toolbarItems.put(i, glassPane);
	        	}
	        }
        }
        
//        // Add the glass bar items.
//        int limit = 1;
//        for (int glassBarItem : glassBarItems.keySet()) {
//        	if (limit >= 9)
//        		break;
//        	inventory.setItem(limit, glassBarItems.get(glassBarItem).getItem());
//        	limit++;
//        }

        // Finally, add the toolbar items.
        for(int toolbarItem : toolbarItems.keySet()){
            int rawSlot = toolbarItem + 36;
            inventory.setItem(rawSlot, toolbarItems.get(toolbarItem).getItem());
        }
        
        return inventory;
    }

    /**
     * Simply an alias to register the Inventory listeners for a certain plugin.
     * Intended to improve code readability.
     *
     * @param plugin The Spigot plugin instance that you wish to register the listeners for.
     */
    public static void prepare(JavaPlugin plugin){
        if(inventoryListenerGUI == null){
            inventoryListenerGUI = new InventoryListenerGUI();
            plugin.getServer().getPluginManager().registerEvents(inventoryListenerGUI, plugin);
        }
    }

}
